﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Common;
using HeuristicLab.Problems.BinPacking.Shapes;
using HeuristicLab.Problems.BinPacking.PackingItem;
using HeuristicLab.Problems.BinPacking.PackingBin;

namespace HeuristicLab.Problems.BinPacking.Dimensions {
  [Item("Occupied Points 3D", "A datastructure holding all currently occupied points of a 3D packing-bin.")]
  [StorableClass]
  public class OccupiedPoints3D : OccupiedPoints<ThreeDimensionalPacking, CuboidPackingBin, CuboidPackingItem> {

    [Storable]
    private int[, ,] occupiedPoints { get; set; }

    public OccupiedPoints3D(CuboidPackingBin binMeasures) : base(binMeasures) {
      occupiedPoints = new int[binMeasures.Width, binMeasures.Height, binMeasures.Depth];
      for (int w = 0; w < binMeasures.Width; w++) {
        for (int h = 0; h < binMeasures.Height; h++) {
          for (int d = 0; d < binMeasures.Depth; d++) {
            occupiedPoints[w, h, d] = -1;
          }
        }
      }
    }
    [StorableConstructor]
    protected OccupiedPoints3D(bool deserializing) : base(deserializing) { }
    protected OccupiedPoints3D(OccupiedPoints3D original, Cloner cloner)
      : base(original, cloner) {
        this.occupiedPoints = (int[,,])original.occupiedPoints.Clone();
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new OccupiedPoints3D(this, cloner);
    }

    public override bool IsPositionFeasible(CuboidPackingItem measures, ThreeDimensionalPacking position) {
      for (int w = position.X; w < position.X + measures.Width; w++) {
        for (int h = position.Y; h < position.Y + measures.Height; h++) {
          for (int d = position.Z; d < position.Z + measures.Depth; d++) {
            if (occupiedPoints[w, h, d] != -1)
              return false;
          }
        }
      }
      return true;
    }
    public override void OccupyPoints(CuboidPackingItem measures, ThreeDimensionalPacking position, int itemID) {
      int width = position.Rotated ? measures.Depth : measures.Width;
      int depth = position.Rotated ? measures.Width : measures.Depth;
      for (int w = position.X; w < position.X + width; w++) {
        for (int h = position.Y; h < position.Y + measures.Height; h++) {
          for (int d = position.Z; d < position.Z + depth; d++) {
            occupiedPoints[w, h, d] = itemID;
          }
        }
      }
    }
    public override bool IsPointOccupied(ThreeDimensionalPacking point) {
      return occupiedPoints[point.X, point.Y, point.Z] != -1;
    }
    public override int ShortestPossibleSideFromEP(ThreeDimensionalPacking ep) {
      int shortestSide = int.MaxValue;
      int width = occupiedPoints.GetLength(0);
      int height = occupiedPoints.GetLength(1);
      int depth = occupiedPoints.GetLength(2);

      if (ep.X >= width || ep.Y >= height || ep.Z >= depth)
        return shortestSide;

      int currentX = ep.X;
      while (currentX < width && occupiedPoints[currentX, ep.Y, ep.Z] == -1) { currentX++; }
      if (currentX - ep.X < shortestSide)
        shortestSide = currentX - ep.X;

      int currentY = ep.Y;
      while (currentY < height && occupiedPoints[ep.X, currentY, ep.Z] == -1) { currentY++; }
      if (currentY - ep.Y < shortestSide)
        shortestSide = currentY - ep.Y;

      int currentZ = ep.Z;
      while (currentZ < depth && occupiedPoints[ep.X, ep.Y, currentZ] == -1) { currentZ++; }
      if (currentZ - ep.Z < shortestSide)
        shortestSide = currentZ - ep.Z;

      return shortestSide;
    }
    public override bool IsStaticStable(CuboidPackingItem item, ThreeDimensionalPacking ep) {
      //Static stability is given, if item is placed on the ground
      if (ep.Y == 0)
        return true;

      if (occupiedPoints[ep.X, ep.Y - 1, ep.Z] != -1
        && occupiedPoints[ep.X + item.Width - 1, ep.Y - 1, ep.Z] != -1
        && occupiedPoints[ep.X, ep.Y - 1, ep.Z + item.Depth - 1] != -1
        && occupiedPoints[ep.X + item.Width - 1, ep.Y - 1, ep.Z + item.Depth - 1] != -1)
        return true;

      //int groundCount = 0;
      //for (int x = ep.X; x < ep.X + item.Width - 1; x++) {
      //  for (int z = ep.Z; z < ep.Z + item.Depth - 1; z++) {
      //    if (occupiedPoints[x,ep.Y-1, z] != -1)                                                     
      //      groundCount++;
      //  }
      //}
      //double stableGround = (double)(groundCount) / (double)(item.Width * item.Depth);
      //if (stableGround > 0.75)
      //  return true;

      return false;
    }
    public override bool WeightIsSupported(CuboidPackingItem item, ThreeDimensionalPacking ep, ItemList<CuboidPackingItem> items) {
      //Static stability is given, if item is placed on the ground
      if (ep.Y == 0)
        return true;

      if (items[occupiedPoints[ep.X, ep.Y - 1, ep.Z]].SupportsStacking(item)
        && items[occupiedPoints[ep.X + item.Width - 1, ep.Y - 1, ep.Z]].SupportsStacking(item)
        && items[occupiedPoints[ep.X, ep.Y - 1, ep.Z + item.Depth - 1]].SupportsStacking(item)
        && items[occupiedPoints[ep.X + item.Width - 1, ep.Y - 1, ep.Z + item.Depth - 1]].SupportsStacking(item))
        return true;

      //int groundCount = 0;
      //for (int x = ep.X; x < ep.X + item.Width - 1; x++) {
      //  for (int z = ep.Z; z < ep.Z + item.Depth - 1; z++) {
      //    if (occupiedPoints[x,ep.Y-1, z] != -1)                                                     
      //      groundCount++;
      //  }
      //}
      //double stableGround = (double)(groundCount) / (double)(item.Width * item.Depth);
      //if (stableGround > 0.75)
      //  return true;

      return false;
    }
    public override void ChangeBinMeasures(CuboidPackingBin binMeasures) {
      int oldDepth = occupiedPoints.GetLength(2);
      var newBinOccupation = new int[binMeasures.Width, binMeasures.Height, binMeasures.Depth];
      for (int w = 0; w < binMeasures.Width; w++) {
        for (int h = 0; h < binMeasures.Height; h++) {
          for (int d = 0; d < binMeasures.Depth; d++) {
            if (d < oldDepth)
              newBinOccupation[w, h, d] = occupiedPoints[w, h, d];
            else
              newBinOccupation[w, h, d] = -1;
          }
        }
      }
      occupiedPoints = newBinOccupation;
    }
  }
}
