﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Problems.BinPacking.PackingItem;
using HeuristicLab.Problems.BinPacking.Shapes;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Common;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using HeuristicLab.Collections;
using HeuristicLab.Problems.BinPacking.Dimensions;
using HeuristicLab.Problems.BinPacking.PackingBin;

namespace HeuristicLab.Encodings.PackingEncoding.PackingPlan {
  [Item("BinPacking3D", "Represents a single-bin packing for a 3D bin-packing problem.")]
  [StorableClass]
  public class BinPacking3D : BinPacking<ThreeDimensionalPacking, CuboidPackingBin, CuboidPackingItem> {

    public BinPacking3D(CuboidPackingBin binMeasures) : base(binMeasures) {
      //OccupiedPoints = new OccupiedPoints3D(binMeasures);
    }
    [StorableConstructor]
    protected BinPacking3D(bool deserializing) : base(deserializing) { }
    protected BinPacking3D(BinPacking3D original, Cloner cloner)
      : base(original, cloner) {
        this.depthWasDoubled = original.depthWasDoubled;
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BinPacking3D(this, cloner);
    }
                                       
    protected override void GenerateNewExtremePointsForNewItem(CuboidPackingItem newItem, ThreeDimensionalPacking position) {
      int newWidth = position.Rotated ? newItem.Depth : newItem.Width;
      int newDepth = position.Rotated ? newItem.Width : newItem.Depth;

      //Find ExtremePoints beginning from sourcepointX
      var sourcePointX = new ThreeDimensionalPacking(0, position.X + newWidth, position.Y, position.Z);
      if (sourcePointX.X < BinMeasures.Width && sourcePointX.Y < BinMeasures.Height && sourcePointX.Z < BinMeasures.Depth) {
        //Traversing down the y-axis                                                                            
        int currentYValue = sourcePointX.Y;
        while (currentYValue > 0 && !IsPointOccupied(new ThreeDimensionalPacking (0, sourcePointX.X, currentYValue, sourcePointX.Z))) {
          currentYValue--;
        }
        ExtremePoints.Add(new ThreeDimensionalPacking(0, sourcePointX.X, currentYValue, sourcePointX.Z));

        //Traversing down the z-axis
        int currentZValue = sourcePointX.Z;
        while (currentZValue > 0 && !IsPointOccupied(new ThreeDimensionalPacking (0, sourcePointX.X, sourcePointX.Y, currentZValue - 1))) {
          currentZValue--;
        }
        ExtremePoints.Add(new ThreeDimensionalPacking(0, sourcePointX.X, sourcePointX.Y, currentZValue));
      }




      //Find ExtremePoints beginning from sourcepointY
      var sourcePointY = new ThreeDimensionalPacking(0, position.X, position.Y + newItem.Height, position.Z);
      if (sourcePointY.X < BinMeasures.Width && sourcePointY.Y < BinMeasures.Height && sourcePointY.Z < BinMeasures.Depth) {
        //Traversing down the x-axis  
        int currentXValue = sourcePointY.X;
        while (currentXValue > 0 && !IsPointOccupied(new ThreeDimensionalPacking (0, currentXValue - 1, sourcePointY.Y, sourcePointY.Z))) {
          currentXValue--;
        }
        ExtremePoints.Add(new ThreeDimensionalPacking(0, currentXValue, sourcePointY.Y, sourcePointY.Z));

        //Traversing down the z-axis
        int currentZValue = sourcePointY.Z;
        while (currentZValue > 0 && !IsPointOccupied(new ThreeDimensionalPacking (0, sourcePointY.X, sourcePointY.Y, currentZValue - 1))) {
          currentZValue--;
        }
        ExtremePoints.Add(new ThreeDimensionalPacking(0, sourcePointY.X, sourcePointY.Y, currentZValue));
      }





      //Find ExtremePoints beginning from sourcepointZ
      var sourcePointZ = new ThreeDimensionalPacking(0, position.X, position.Y, position.Z + newDepth);
      if (sourcePointZ.X < BinMeasures.Width && sourcePointZ.Y < BinMeasures.Height && sourcePointZ.Z < BinMeasures.Depth) {
        //Traversing down the x-axis  
        int currentXValue = sourcePointZ.X;
        while (currentXValue > 0 && !IsPointOccupied(new ThreeDimensionalPacking (0, currentXValue - 1, sourcePointZ.Y, sourcePointZ.Z))) {
          currentXValue--;
        }
        ExtremePoints.Add(new ThreeDimensionalPacking(0, currentXValue, sourcePointZ.Y, sourcePointZ.Z));

        //Traversing down the y-axis                                                                            
        int currentYValue = sourcePointZ.Y;
        while (currentYValue > 0 && !IsPointOccupied(new ThreeDimensionalPacking (0, sourcePointZ.X, currentYValue, sourcePointZ.Z))) {
          currentYValue--;
        }
        ExtremePoints.Add(new ThreeDimensionalPacking(0, sourcePointZ.X, currentYValue, sourcePointZ.Z));
      }

      ExtremePoints = new HashSet<ThreeDimensionalPacking>(ExtremePoints.
        OrderBy(ep => ep.Z).
        ThenBy(ep => ep.X).
        ThenBy(ep => ep.Y).
        ThenBy(ep => ShortestPossibleSideFromPoint(ep)));
    }

    public override ThreeDimensionalPacking FindExtremePointForItem(CuboidPackingItem measures, bool rotated, bool stackingConstraints) {

      CuboidPackingItem item = new CuboidPackingItem(
        rotated ? measures.Depth : measures.Width,
        measures.Height,
        rotated ? measures.Width : measures.Depth,
        measures.TargetBin);

      int epIndex = 0;
      while (epIndex < ExtremePoints.Count && (
        !IsPositionFeasible(item, ExtremePoints.ElementAt(epIndex)) || (stackingConstraints && !IsStaticStable(item, ExtremePoints.ElementAt(epIndex)))
      )) { epIndex++; }

      if (epIndex < ExtremePoints.Count) {
        var result = ExtremePoints.ElementAt(epIndex);
        result.Rotated = rotated;
        return result;
      }
      return null;
    }
    public override ThreeDimensionalPacking FindPositionBySliding(CuboidPackingItem measures, bool rotated) {
      //Starting-position at upper right corner (=left bottom point of item-rectangle is at position item.width,item.height)
      ThreeDimensionalPacking currentPosition = new ThreeDimensionalPacking(0,
        BinMeasures.Width - (rotated ? measures.Depth : measures.Width),
        BinMeasures.Height - measures.Height,
        BinMeasures.Depth - (rotated ? measures.Width : measures.Depth), rotated);
      //Slide the item as far as possible to the bottom
      while (IsPositionFeasible(measures, ThreeDimensionalPacking.MoveDown(currentPosition))
        || IsPositionFeasible(measures, ThreeDimensionalPacking.MoveLeft(currentPosition))
        || IsPositionFeasible(measures, ThreeDimensionalPacking.MoveBack(currentPosition))) {
        //Slide the item as far as possible to the left
          while (IsPositionFeasible(measures, ThreeDimensionalPacking.MoveLeft(currentPosition))
        || IsPositionFeasible(measures, ThreeDimensionalPacking.MoveBack(currentPosition))) {
          //Slide the item as far as possible to the back
          while (IsPositionFeasible(measures, ThreeDimensionalPacking.MoveBack(currentPosition))) {
            currentPosition = ThreeDimensionalPacking.MoveBack(currentPosition);
          }
          if (IsPositionFeasible(measures, ThreeDimensionalPacking.MoveLeft(currentPosition)))
            currentPosition = ThreeDimensionalPacking.MoveLeft(currentPosition);
        }
          if (IsPositionFeasible(measures, ThreeDimensionalPacking.MoveDown(currentPosition)))
            currentPosition = ThreeDimensionalPacking.MoveDown(currentPosition);
      }

      return IsPositionFeasible(measures, currentPosition) ? currentPosition : null;
    }
    
    public override void SlidingBasedPacking(ref List<int> sequence, ItemList<CuboidPackingItem> itemMeasures) {
      var temp = new List<int>(sequence);
      for (int i = 0; i < temp.Count; i++) {
        var item = itemMeasures[temp[i]];
        var position = FindPositionBySliding(item, false);
        if (position != null) {
          PackItem(temp[i], item, position);
          sequence.Remove(temp[i]);
        }
      }
    }
    public override void SlidingBasedPacking(ref List<int> sequence, ItemList<CuboidPackingItem> itemMeasures, Dictionary<int, bool> rotationArray) {
      var temp = new List<int>(sequence);
      for (int i = 0; i < temp.Count; i++) {
        var item = itemMeasures[temp[i]];
        var position = FindPositionBySliding(item, rotationArray[i]);
        if (position != null) {
          PackItem(temp[i], item, position);
          sequence.Remove(temp[i]);
        }
      }
    }
    public override void ExtremePointBasedPacking(ref List<int> sequence, ItemList<CuboidPackingItem> itemMeasures, bool stackingConstraints) {
      var temp = new List<int>(sequence);
      foreach (int itemID in temp) {
        var item = itemMeasures[itemID];
        var positionFound = FindExtremePointForItem(item, false, stackingConstraints);
        if (positionFound != null) {
          PackItem(itemID, item, positionFound);
          sequence.Remove(itemID);
        }
      }
    }
    public override void ExtremePointBasedPacking(ref List<int> sequence, ItemList<CuboidPackingItem> itemMeasures, bool stackingConstraints, Dictionary<int, bool> rotationArray) {
      var temp = new List<int>(sequence);
      foreach (int itemID in temp) {
        var item = itemMeasures[itemID];
        var positionFound = FindExtremePointForItem(item, rotationArray[itemID], stackingConstraints);
        if (positionFound != null) {
          PackItem(itemID, item, positionFound);
          sequence.Remove(itemID);
        }
      }
    }
    

    [Storable]
    private bool depthWasDoubled = false;
    public void DoubleDepth() {
      if (!depthWasDoubled) {
        var oldDepth = BinMeasures.Depth;
        BinMeasures.Depth = BinMeasures.Depth * 2;
        //OccupiedPoints.ChangeBinMeasures(BinMeasures);
        depthWasDoubled = true;
      }
    }

    public override int ShortestPossibleSideFromPoint(ThreeDimensionalPacking position) {

      int shortestSide = int.MaxValue;
      int width = BinMeasures.Width;
      int height = BinMeasures.Height;
      int depth = BinMeasures.Depth;

      if (position.X >= width || position.Y >= height || position.Z >= depth)
        return shortestSide;

      ThreeDimensionalPacking current = new ThreeDimensionalPacking (0, position.X, position.Y, position.Z);
      while (current.X < width && IsPointOccupied(current)) { current = ThreeDimensionalPacking.MoveRight(current); }
      if (current.X - position.X < shortestSide)
        shortestSide = current.X - position.X;


      current = new ThreeDimensionalPacking(0, position.X, position.Y, position.Z);
      while (current.Y < height && IsPointOccupied(current)) { current = ThreeDimensionalPacking.MoveUp (current); }
      if (current.Y - position.Y < shortestSide)
        shortestSide = current.Y - position.Y;


      current = new ThreeDimensionalPacking(0, position.X, position.Y, position.Z);
      while (current.Z < depth && IsPointOccupied(current)) { current = ThreeDimensionalPacking.MoveFront(current); }
      if (current.Z - position.Z < shortestSide)
        shortestSide = current.Z - position.Z;

      return shortestSide;
    }
    public override bool IsStaticStable(CuboidPackingItem item, ThreeDimensionalPacking position) {
      //Static stability is given, if item is placed on the ground
      if (position.Y == 0)
        return true;

      if (IsPointOccupied (new ThreeDimensionalPacking (0, position.X, position.Y - 1, position.Z))
        && IsPointOccupied (new ThreeDimensionalPacking (0, position.X + item.Width - 1, position.Y - 1, position.Z))
        && IsPointOccupied (new ThreeDimensionalPacking (0, position.X, position.Y - 1, position.Z + item.Depth - 1))
        && IsPointOccupied (new ThreeDimensionalPacking (0, position.X + item.Width - 1, position.Y - 1, position.Z + item.Depth - 1)))
        return true;


      //if (occupiedPoints[position.X, position.Y - 1, position.Z] != -1
      //  && occupiedPoints[position.X + item.Width - 1, position.Y - 1, position.Z] != -1
      //  && occupiedPoints[position.X, position.Y - 1, position.Z + item.Depth - 1] != -1
      //  && occupiedPoints[position.X + item.Width - 1, position.Y - 1, position.Z + item.Depth - 1] != -1)
      //  return true;

      //int groundCount = 0;
      //for (int x = ep.X; x < ep.X + item.Width - 1; x++) {
      //  for (int z = ep.Z; z < ep.Z + item.Depth - 1; z++) {
      //    if (occupiedPoints[x,ep.Y-1, z] != -1)                                                     
      //      groundCount++;
      //  }
      //}
      //double stableGround = (double)(groundCount) / (double)(item.Width * item.Depth);
      //if (stableGround > 0.75)
      //  return true;

      return false;
    }
  }
}
